/*********************************************/
/* INTERFACE CLASSES TO Pythia8              */
/* feed pythia with a completely new process */
/*********************************************/
//-----------------
#include "pythiaPlugin.h"

using namespace std;
using namespace Pythia8; 

namespace Moses1 {

// A derived class for a completely new, user defined (from scratch) process, given as an input to pythia

void KaluzaKlein::initProc() {

	// initialize parameters
	m_prm.setParameters(m_physPrm); // initialize the new instance of the physical parameters struct
	m_prm.setParameters(m_runPrm); // initialize the new instance of the run parameters struct
	m_prm.printUnits(); // just a reminder about units...

	cout << "\n------------- m_runPrm.invMassCut before = " << m_runPrm.invMassCut << " -------------" << endl;
	m_runPrm.invMassCut = settingsPtr->parm("PhaseSpace:mHatMin");
	cout << "------------- m_runPrm.invMassCut after = " << m_runPrm.invMassCut << " -------------\n" << endl; 
	
	// initialize
	m_pdtPrt.setPDTtable("HEPPDTTBL","particle.tbl"); // can also initialize with full path and name

	// initialize an arbitrary array of id's of the leptons/quarks to be loaded. using "pdtParticle" enum
	// d, u, s, c, b, t, e, mu, tau, nu_e, nu_mu, nu_tau
	int idFermions[] = {PDTDWN,PDTUP,PDTSTR,PDTCHM,PDTBOT,PDTTOP,PDTE,PDTMU,PDTTAU,PDTNUE,PDTNUMU,PDTNUTAU};
	// fill all quarks and leptons properties in a single vector:
	m_pdtPrt.getAll(idFermions, sizeof(idFermions)/sizeof(*idFermions), m_PDTfermions, false);

	// do the same for PDTbosons:
	int idBosons[] = {PDTGLU,PDTGAMMA,PDTZ,PDTWPLUS}; // glu, gamma, Z0, W+
	m_pdtPrt.getAll(idBosons, sizeof(idBosons)/sizeof(*idBosons), m_PDTbosons, false); // fill all PDTbosons properties in a single vector

	// add to the previously created vector of PDT particle 2 more properties to each properties struct:
	// 1) weakI3, 2) Nc (Number of possible colors)
	// see: mosesParticle.cxx for more details of how this is actually being done
	// put only the selected pdtParticle structs from the PDTfermions vector in the CUSTOMparticle and add more properties:
	m_customPrt.setCUSTOMparticle(m_PDTfermions, m_CUSTOMfermions);
	// from now on, have to use the new vector in order to get the custom properties.
	// the PDT vector also exist and can be used.
	// note that the bosons vector is not modified here

	// initialize lha (lhapdf instance)
	//string pdfSet = "MRST2001lo.LHgrid";
	//string pdfSet = "cteq6.LHpdf";
	//string pdfSet = "cteq5l.LHgrid";
	//string pdfSet = "MRST2006nnlo.LHgrid";
	string pdfSet = m_pdfSet;
	int subset = m_pdfMember;
	m_lha.setlhapdf("LHAPDFSETS",pdfSet,subset); // can also initialize with full path and name

	// the selection of the outgoing fermion is done in the main program where m_idNew is being initialized
	for (int i=0 ; i<(int)m_CUSTOMfermions.size() ; i++) {
		if ( m_CUSTOMfermions[i].pdt_prt.id == (int)abs(m_idNew) ) {m_outFermion = m_CUSTOMfermions[i];}
	}
	cout << "outFermions are: " << m_outFermion.pdt_prt.name << " and its anti" << endl;
	// the selection of the number of KK modes is done in the main program where m_nKKmodes is being initialized
	cout << "number of KK modes is: " << m_nKKmodes << endl;

	// Secondary open width fraction, relevant for top (or heavier).
	m_openFracPair = particleDataPtr->resOpenFrac(m_idNew, -m_idNew);	

	cout << "initProc()...OK" << endl;
}


// Evaluate sigmaKin(sHat); first step when inflavours unknown. 
void KaluzaKlein::sigmaKin() {
	
	// Check that above threshold.
	m_isPhysical = true;
	if (mH < m3 + m4 + MASSMARGIN) {
		m_isPhysical = false;
		return;
	}
	// Define average F, Fbar mass so same beta. Phase space.
	double s34Avg  = 0.5 * (s3 + s4) - 0.25 * pow2(s3 - s4) / sH;
	m_mr             = s34Avg / sH;
	m_betaf          = sqrtpos(1. - 4. * m_mr);

	// Reconstruct decay angle so can reuse 2 -> 1 cross section.
	m_cosThe         = (tH - uH) / (m_betaf * sH);

	/*
	m_totalwidth_Z0 = 0;
	m_totalwidth_gammaKK = 0;
	m_totalwidth_ZKK = 0;
	// calculate the total width of Z0 for each s_hat:
	for (int i=0 ; i<(int)m_CUSTOMfermions.size() ; i++) {
		switch (m_CUSTOMfermions[i].pdt_prt.id) {
			case PDTNUE: break;
			case PDTNUMU: break;
			case PDTNUTAU: break;
			default: // gamma can not decay into neutrino pair (while Z can)
				//m_totalwidth_gammaKK += m_wdt.getpartialWidth(1,sH,m_PDTbosons[1],m_CUSTOMfermions[i],m_physPrm); // gammaStar width
				break;
		}
		//m_totalwidth_ZKK += m_wdt.getpartialWidth(1, sH, m_PDTbosons[2], m_CUSTOMfermions[i], m_physPrm); // Z* total width 
		m_totalwidth_Z0 += m_wdt.getpartialWidth(0, sH, m_PDTbosons[2], m_CUSTOMfermions[i], m_physPrm); // Z0 total width
	}
	*/
	m_totalwidth_Z0 = m_PDTbosons[2].width;
}



// Evaluate sigmaHat(sHat); first step when inflavours unknown. 
double KaluzaKlein::sigmaHat() {

        // Fail if below threshold.
        if (!m_isPhysical) {cout << "not physical" << endl; return 0.;}
        // Cross section vanishing except for incoming qqar.
        if (abs(id1) != abs(id2)) {cout << "id1!=-id2" << endl; return 0.;}
        // if not supported quarks
        if ( (int)abs(id1)<PDTDWN || (int)abs(id1)>PDTTOP ) {cout << "parton id=" << (int)abs(id1) << " is not supported" << endl; return 0.;}
        // if not supported final state 1
        if ( (int)abs(m_idNew)<PDTE && (int)abs(id1)>=PDTTOP ) {cout << "final state id=" << (int)abs(m_idNew)
								     << " is not supported" << endl; return 0.;}
        // if not supported final state 2
        if ( (int)abs(m_idNew)<PDTDWN || (int)abs(m_idNew)>PDTNUTAU ) {cout << "final state id="
									    << (int)abs(m_idNew) << " is not supported" << endl; return 0.;}

        for (int i=0 ; i<(int)m_CUSTOMfermions.size() ; i++) {
                if ( m_CUSTOMfermions[i].pdt_prt.id == (int)abs(id1) ) {m_inFermion = m_CUSTOMfermions[i];}
        }

        // initialize values
        m_sigma = 0;
        // calculate the averaged (over all helicity states) helicity square amplitude while summing all possible polarizations
        for (double hIn=-0.5 ; hIn<=0.5 ; hIn++) { // loop over incoming helicity states (all quarks)
                for (double hOut=-0.5 ; hOut<=0.5 ; hOut++) { // loop over outgoing helicity states (muon only)

                        // get the angularDistribution (determined by helicity states)
                        m_thetaDistribution = 0;
                        if (hIn==hOut) {m_thetaDistribution = (1+m_cosThe)*(1+m_cosThe);}
                        else {m_thetaDistribution = (1-m_cosThe)*(1-m_cosThe);}

			m_fullAmp = dcomplex(0,0);
                        // calculate square amplitude for each helicity states combinations
			// ---------- SM terms ----------
			// photon propagator
			m_fullAmp += m_hAmp.gethelicityAmplitude(sH,0,hIn,hOut,m_inFermion,m_outFermion,m_PDTbosons[1],m_physPrm);
			// Z0 propagator
			m_fullAmp += m_hAmp.gethelicityAmplitude(sH,m_totalwidth_Z0,hIn,hOut,m_inFermion,m_outFermion,m_PDTbosons[2],m_physPrm);
			// ------------------------------

			// ---------- KK terms ----------
			for (int KKmode=1 ; KKmode<=m_nKKmodes ; KKmode++) { // loop over all KK modes
				
				// calculate the width for every excitation
				m_totalwidth_gammaKK = 0;
				m_totalwidth_ZKK = 0;
				// calculate the total width of Z0 for each s_hat:
				for (int i=0 ; i<(int)m_CUSTOMfermions.size() ; i++) {
					switch (m_CUSTOMfermions[i].pdt_prt.id) {
						case PDTNUE: break;
						case PDTNUMU: break;
						case PDTNUTAU: break;
						default: // gamma can not decay into neutrino pair (while Z can)
							m_totalwidth_gammaKK += m_wdt.getpartialWidth(KKmode,sH,m_PDTbosons[1],m_CUSTOMfermions[i],m_physPrm); // gammaStar width
							break;
					}
					m_totalwidth_ZKK += m_wdt.getpartialWidth(KKmode, sH, m_PDTbosons[2], m_CUSTOMfermions[i], m_physPrm); // Z* total width 
				}
			
				// photon* propagator
				m_fullAmp +=  m_hAmp.gethelicityAmplitude(KKmode, sH,
									m_totalwidth_gammaKK,
									hIn, hOut,
									m_inFermion, m_outFermion,
									m_PDTbosons[1], m_physPrm);

				// Z* propagator
				m_fullAmp +=  m_hAmp.gethelicityAmplitude(KKmode, sH,
									m_totalwidth_ZKK,
									hIn, hOut,
									m_inFermion, m_outFermion,
									m_PDTbosons[2], m_physPrm);
			}
			//-------------------------------

			m_fullAmp *= conj(m_fullAmp);
			m_sigma += (1./4.) * ( real(m_fullAmp) * m_thetaDistribution );
		} // end hIn
	} // end hOut

	// Top: corrections for closed decay channels.
	m_sigma *= m_openFracPair;

	// Multiply the averaged squared helicity amplitude by some factors:
	// [1] 2/sHat is received from the transformation between dtHat and dcosTheta.
	//     A 2->1 process such as qqbar->gamma*/Z0 is only sigmaHat(sHat).
	//     If you choose to include the decay, using a 2->2 process (2->1->2),
	//     then you instead use d(sigmaHat)/d(tHat), which dimensionally is a factor
	//     1/sHat different. Althernatively you can think of it in terms of the phase
	//     space integral 1 = integral_{-sHat}{0} d{tHat)/sHat.
	//     tHat=(-sHat/2)*(1-cosTheta) ==> dtHat=(sHat/2)*dcosTheta ==> d/dtHat=(2/sHat)*d/dcosTheta.
	// [2] 2*pi is the integral over dphi
	// [3] (alpha^2/4sHat)*(Nc_f/Nc_q) is the normal outer factors out of the averaged matrix element
	// [4] sHat^2 is to fix the units and to cancel the 1/sHat^2 that is artificially included in the
	//     matrix element (for the photon bit it is inserted by hand, for the Z0 bit it is inherent)
	m_sigma *= (2./sH) * (2.*m_physPrm.pi)
		   * pow2(alpEM)/(4.*sH) * (double)m_outFermion.Nc/(double)m_inFermion.Nc // use running alpha_em
		   * pow2(sH);
	
		// [m_sigma] = [1/GeV^4] --> [Integrate{m_sigma*ds}] = [1/GeV^2] --> OK !!!
		/*--------------------------------------------------------------------------------*/
		/* it is assumed that cross sections normally come in dimensions such that they,  */
		/* when integrated over the relevant phase space, obtain the dimension [1/GeV^2], */
		/* and therefore need to be converted to mb                                       */
		/* If the cross section is already encoded as mb then convert2mb() should be      */
		/* overloaded to instead return false                                             */
		/*--------------------------------------------------------------------------------*/

	// incoming color possibilities (PDFs returns 9 combinations, only 3 are valid for neutral current)
	if (abs(id1)<9) {m_sigma /= 3;}

	// Cross section as already calculated.
	return m_sigma;

}



// Select identity, colour and anticolour.
void KaluzaKlein::setIdColAcol() {

	// Set outgoing flavours.
	id3 = (id1 > 0) ? m_idNew : -m_idNew;
	setId( id1, id2, id3, -id3);


	//setId(id1, id2, idOut, -1*idOut); // id1,id2 is given by pythia

	// Colour flow topologies. Swap when antiquarks.
	if (id1<0) {setColAcol(0,1, 1,0, 0,0, 0,0);} // id1 is an anti-quark
	else {setColAcol(1,0, 0,1, 0,0, 0,0);} // id1 is a quark 
}



}
